#include <arduino.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "sseg.h"


// Definition constance

//#define DEBUG

#define BUTTON_DOWN 0b00000001
#define BUTTON_UP   0b00000010
#define BUTTON_F0   0b00000100
#define BUTTON_F1   0b00001000
#define BUTTON_F2   0b00010000
#define BUTTON_DIR  0b00100000

#define MODE_CV			1
#define MODE_LOC		2
#define MODE_INIT		3

#define PIN_PWM_A		9
#define PIN_PWM_B		10

#define WAIT_MSG 480

#define MIN_LOCADDRESS	1
#define MAX_LOCADDRESS	9999
#define MIN_ACCADDRESS	1
#define MAX_ACCADDRESS	999


// Definition for variables
DSCoreLib Gateway;
SSEG SegDisplay(6, 5, 1, 0, A5, A3, 2, 3, 4, A4, A2);


//Loc mode
word gLocAddress = 3;
word eepLocAddress = 3;
byte gLocDirection = GO_FWD;
byte gLocSpeed = 0;

byte gButton_val = 0;
byte gLocFunc0 = 0;
byte gLocFunc1 = 0;
byte gLocFunc2 = 0;

word gDisplayWordCount = 0;
byte gDisplayNext[4] = {SSEG_NONE, SSEG_NONE, SSEG_NONE, SSEG_NONE};

byte gMode_Main = 0;
byte gMode_Drive = 0;
byte gLCDCounter = 0;

//Last Speed
word gLastSpeed = 9999;

//Protcol
word gLocProtcol = ADDR_DCC;
word gAccProtcol = ADDR_ACC_DCC - 1;


//Acc mode
word gAccAddress = 1;


//CV mode
byte gMode_CV = 0;
byte gCVAddress = 1;
byte gCVValue = 0;


unsigned long gTime_5ms = 0;
unsigned long gTime_100ms = 0;
byte gDisplayCounter = 0;

word gPressCount = 0; /* */


// Functions
void displayModeCV();
void displayModeLoc();
void displayModeAcc();
void displayTemporary();
void processModeCV();
void processModeDial();
void processModeLoc();
byte getButtonStatus();
bool checkButton(byte inBuf, byte inCurrent, byte inButtonBit);
bool checkButtonLeaved(byte inBuf, byte inCurrent, byte inButtonBit);
bool checkButtonContinuous(byte inBuf, byte inCurrent, byte inButtonBit);
void saveEEPROM(byte inMode);
void loadEEPROM();
void updateEEPROM();

/*******************************
    Setup
********************************/

void setup()
{
	
	//Load EEPROM
	loadEEPROM();
	
	//Analog input
	pinMode(A0, INPUT);
	digitalWrite(A0, LOW); // disable internal pullup
	analogReference(DEFAULT); // set analog reference 5V(VCC) 
	
	//Faster Analog read
	ADCSRA = ADCSRA & 0xf8;
	ADCSRA = ADCSRA | 0x04;
	
	//Button
	pinMode(8, INPUT);
	pinMode(7, INPUT);
	pinMode(11, INPUT);
	pinMode(12, INPUT);
	pinMode(13, INPUT);
	pinMode(A1, INPUT);
	digitalWrite(8, HIGH);
	digitalWrite(7, HIGH);
	digitalWrite(11, HIGH);
	digitalWrite(12, HIGH);
	digitalWrite(13, HIGH);
	digitalWrite(A1, HIGH);
	
	pinMode(0, OUTPUT);
	pinMode(1, OUTPUT);

	
	
	//Set default button status
	gButton_val = getButtonStatus();
	
	//CV mode check
	
	if( ((gButton_val & BUTTON_F0) | (gButton_val & BUTTON_F1) | (gButton_val & BUTTON_F2)) > 0)
	{
		gMode_Main = MODE_CV;
		
	}
	else
	{
		gMode_Main = MODE_INIT;
		
	}
	
	
	//Init Gateway
	Gateway.Init();
	
	//LCD
	
	switch( gMode_Main)
	{
	case MODE_CV:
		SegDisplay.SetCVNumber(gCVAddress);
		break;
		
	case MODE_INIT:
		SegDisplay.SetWord(SSEG_M, SSEG_I, SSEG_N, SSEG_T);
		break;
	}
	
	//^C}(7ZOp)
	TCCR1A = 0; // 
	TCCR1B = 0; // 
	OCR1A = 165;// 400Hz(2.5ms)
	TCCR1B |= (1 << CS12) | (1 << WGM12);      // 256, CTC[hɐݒ
	TIMSK1 |= (1 << OCIE1A); 
	sei();
	//^C}I
	
}

/*******************************
	400Hz(2.5ms) Interval Interrupt
********************************/


ISR (TIMER1_COMPA_vect)
{
	//1 word showing (4 word as 10ms)
	process7Seg();
}



/*******************************
    Main loop
********************************/

void loop()
{
	//Scan
	if( (gLocAddress != 0) && (gMode_Main != MODE_CV))
	{
		//Power on
		if( Gateway.IsPower() == false)
		{
			Gateway.SetPower(true);
		}
		Gateway.Scan();
	}
	
	
	// 100ms interval
	if( (millis() - gTime_100ms) >= 100)
	{
		switch( gMode_Main)
		{
		case MODE_INIT:
			
			if( gLCDCounter > 10 )
			{
				SegDisplay.SetNumber(gLocAddress);
				gMode_Main = MODE_LOC;
			}
			else
			{
				gLCDCounter++;
			}
			
			break;
			
		case MODE_CV:
			
			processModeCV();
			
			break;
			
		case MODE_LOC:
			
			//Update dial for speed control
			processModeDial();
			
			
			if( gMode_Drive == 0)
			{
				processModeLoc();
			}
			else
			{
				processModeAcc();
			}			
			
			break;
		}
		
		//Reset task
		gTime_100ms = millis();
	}
}

/*******************************
    Segment Process
********************************/

void process7Seg(void)
{
	
	//2ms interval
	if( (millis() - gTime_5ms) >= 2)
	{
		//Reset task
		gTime_5ms = millis();
		
		switch( gMode_Main)
		{
		case MODE_INIT:
			SegDisplay.ShowWord();
			break;

		case MODE_CV:
			
			displayModeCV();
			
			break;
			
		case MODE_LOC:
			
			if( gMode_Drive == 0)
			{
				displayModeLoc();
			}
			else
			{
				displayModeAcc();
			}
			
			
			break;
		}
	}	
}


/*******************************
    DisplayModeCV
********************************/

void displayModeCV()
{
	//LCD control
	SegDisplay.ShowWord();
}

/*******************************
    DisplayModeLoc
********************************/

void displayModeLoc()
{
	//LCD control
	if( gDisplayWordCount > 0)
	{
		displayTemporary();
	}
	else
	{
		
		SegDisplay.ShowNumber();
	}
}

/*******************************
    DisplayModeAcc
********************************/

void displayModeAcc()
{
	//LCD control
	if( gDisplayWordCount > 0)
	{
		displayTemporary();
	}
	else
	{
		SegDisplay.ShowWord();
	}
}

/*******************************
    DisplayTemporary
********************************/

void displayTemporary()
{
	//LCD control
	SegDisplay.ShowWord();
	gDisplayWordCount--;
	
	//Twice message function
	if( gDisplayWordCount <= 0)
	{
		
		switch( gMode_Main)
		{
		case MODE_INIT:
			break;
			
		case MODE_CV:
			
			break;
			
		case MODE_LOC:
			
			
			if( gMode_Drive == 0)
			{
				SegDisplay.SetNumber(gLocAddress);
			}
			else
			{
				SegDisplay.SetAccessory(gAccAddress);
			}
			
			break;
		}
	}
}

/*******************************
    processModeCV
********************************/

void processModeCV()
{
	
	//Buttons
	byte aButton_val = getButtonStatus();
	
	if(aButton_val != gButton_val)
	{
		
		
		if(checkButton(gButton_val, aButton_val, BUTTON_UP))
		{
			
			switch( gMode_CV)
			{
			case 0:
				//CV No select
				//Increment CV no
				gCVAddress++;
				
				if(gCVAddress >= 100)
				{
					gCVAddress = 1;
				}
				
				//Display CV No.
				SegDisplay.SetCVNumber(gCVAddress);
				
				break;
				
			case 1:
				//CV Value select
				//Increment CV no
				gCVValue++;
				
				if(gCVValue >= 256)
				{
					gCVValue = 0;
				}
				
				//Display CV Val.
				SegDisplay.SetCVValue(gCVValue);
				break;
			}
			
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_DOWN))
		{
			
			switch( gMode_CV)
			{
			case 0:
				//CV No select
				//Decrement CV no
				
				if(gCVAddress == 1)
				{
					gCVAddress = 99;
				}
				else
				{
					//Decrement address
					gCVAddress--;
				}
				
				//Display CV No.
				SegDisplay.SetCVNumber(gCVAddress);
				break;
				
			case 1:
				//CV Value select
				//Decrement CV no
				
				if(gCVValue == 0)
				{
					gCVValue = 255;
				}
				else
				{
					//Decrement address
					gCVValue--;
				}
				
				//Display CV Val.
				SegDisplay.SetCVValue(gCVValue);
				break;
				
			}
			
		}
		else if (checkButton(gButton_val, aButton_val, BUTTON_F0))
		{
			//[hؑ
			gMode_CV++;
			
			switch( gMode_CV)
			{
			case 0:
				
				break;
				
			case 1:
				
				//Display CV Value.
				SegDisplay.SetCVValue(gCVValue);
				break;
				
			case 2:
				//CV Write
				Gateway.WriteConfig_Dir(gCVAddress, gCVValue);
				
				//Display Loc No.
				SegDisplay.SetCVNumber(gCVAddress);
				
				//Mode
				gMode_CV = 0;
				break;
				
			}
			
			//DEBUG
			
		}
		else if (checkButton(gButton_val, aButton_val, BUTTON_F2))
		{
			
			if( gMode_CV == 0)
			{
				//
			}
			else
			{
				//[hؑ
				gMode_CV--;
			}
			
			
			if( gMode_CV == 0)
			{
				//Display CV No.
				SegDisplay.SetCVNumber(gCVAddress);
			}
			
		}
		
		
		
		//Store to buf
		gButton_val = aButton_val;
		
	}
	else
	{
	}
	
}


/*******************************
    processModeDial
********************************/

void processModeDial()
{
	//Dial
	word aSpeed = (analogRead(A0) & 0b1111111100);
	
	if( ((aSpeed & 0b1111111100) != gLocSpeed))
	{
		gLocSpeed = aSpeed & 0b1111111100;
		
		Gateway.SetLocoSpeedEx(gLocProtcol + gLocAddress, aSpeed, 2);
		
		//Update EEPROM
		updateEEPROM();
		
		//Debug
		if( gLastSpeed != aSpeed)
		{
			gLastSpeed = aSpeed;
		}
	}
}

/*******************************
    processModeLoc
********************************/

void processModeLoc()
{
	//Buttons
	byte aButton_val = getButtonStatus();
	
	if(aButton_val != gButton_val)
	{
		
		
		if(checkButton(gButton_val, aButton_val, BUTTON_UP))
		{
			//Increment address
			gLocAddress++;
			
			if(gLocAddress > MAX_LOCADDRESS)
			{
				gLocAddress = MIN_LOCADDRESS;
			}
			
			//Display Loc No.
			SegDisplay.SetNumber(gLocAddress);
			
			//Reset press count
			gPressCount = 0;
			
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_DOWN))
		{
			
			if(gLocAddress == MIN_LOCADDRESS)
			{
				gLocAddress = MAX_LOCADDRESS;
			}
			else
			{
				//Decrement address
				gLocAddress--;
			}
			
			//Display Loc No.
			SegDisplay.SetNumber(gLocAddress);
			
			//Reset press count
			gPressCount = 0;
			
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F0))
		{
			
			//Reset press count
			gPressCount = 0;
			
		}
		else if(checkButtonLeaved(gButton_val, aButton_val, BUTTON_F0))
		{
			if( gLocAddress == 0)
			{
				return;
			}
			
			//DCC mode
			
			gLocFunc0 = (gLocFunc0 == 0) ? 1 : 0;
			
			Gateway.SetLocoFunction(gLocProtcol + gLocAddress, 1, gLocFunc0);
			
			//Reset press count
			gPressCount = 0;

			//Display Message
			SegDisplay.SetWord(SSEG_F, SSEG_0, SSEG_O, (gLocFunc0 == 0) ? SSEG_F : SSEG_N);
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
			
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F1))
		{
			if( gLocAddress == 0)
			{
				return;
			}
			
			gLocFunc1 = (gLocFunc1 == 0) ? 1 : 0;
			
			Gateway.SetLocoFunction(gLocProtcol + gLocAddress, 2, gLocFunc1);
			
			//Display Message
			SegDisplay.SetWord(SSEG_F, SSEG_1, SSEG_O, (gLocFunc1 == 0) ? SSEG_F : SSEG_N);
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F2))
		{
			if( gLocAddress == 0)
			{
				return;
			}
			
			gLocFunc2 = (gLocFunc2 == 0) ? 1 : 0;
			
			Gateway.SetLocoFunction(gLocProtcol + gLocAddress, 3, gLocFunc2);
			
			//Display Message
			SegDisplay.SetWord(SSEG_F, SSEG_2, SSEG_O, (gLocFunc2 == 0) ? SSEG_F : SSEG_N);
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_DIR))
		{
			
			gLocDirection = (gLocDirection == GO_FWD) ? GO_REV : GO_FWD;
			
			if( gLocAddress == 0)
			{
				return;
			}
			
			
			Gateway.SetLocoDirection(gLocProtcol + gLocAddress, gLocDirection);
			
			//Display Message
			if( gLocDirection == GO_FWD)
			{
				SegDisplay.SetWord(SSEG_NONE, SSEG_F, SSEG_W, SSEG_D);
			}
			else
			{
				SegDisplay.SetWord(SSEG_NONE, SSEG_R, SSEG_E, SSEG_V);
			}
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
		}
		
		
		
		//Store to buf
		gButton_val = aButton_val;
		
	}
	else
	{
		//Buttons have no difference from previous state
		
		if(checkButtonContinuous(gButton_val, aButton_val, BUTTON_UP))
		{
			if( gPressCount > 15)
			{
				//Increment address
				gLocAddress++;
				
				if(gLocAddress > MAX_LOCADDRESS)
				{
					gLocAddress = MIN_LOCADDRESS;
				}
				
				//Display Loc No.
				SegDisplay.SetNumber(gLocAddress);
			}
			else
			{
				gPressCount++;
			}
		}
		else if(checkButtonContinuous(gButton_val, aButton_val, BUTTON_DOWN))
		{
			if( gPressCount > 15)
			{
				if(gLocAddress == MIN_LOCADDRESS)
				{
					gLocAddress = MAX_LOCADDRESS;
				}
				else
				{
					//Decrement address
					gLocAddress--;
				}
				
				//Display Loc No.
				SegDisplay.SetNumber(gLocAddress);
			}
			else
			{
				gPressCount++;
			}
		}
		
		else if(checkButtonContinuous(gButton_val, aButton_val, BUTTON_F0))
		{
			if( gPressCount > 15)
			{
				gMode_Drive = 1;
				
				//Display Acc No.
				SegDisplay.SetAccessory(gAccAddress);
			}
			else
			{
				if(gLocAddress != 0)
				{
					gPressCount++;
				}
				
			}
		}		
		
	}
	
}

/*******************************
    processModeAcc
********************************/

void processModeAcc()
{
	//Buttons
	byte aButton_val = getButtonStatus();
	
	if(aButton_val != gButton_val)
	{
		
		
		if(checkButton(gButton_val, aButton_val, BUTTON_UP))
		{
			//Increment address
			gAccAddress++;
			
			if(gAccAddress > MAX_ACCADDRESS)
			{
				gAccAddress = MIN_ACCADDRESS;
			}
			
			//Display Acc No.
			SegDisplay.SetAccessory(gAccAddress);
		
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_DOWN))
		{
			
			if(gAccAddress == MIN_ACCADDRESS)
			{
				gAccAddress = MAX_ACCADDRESS;
			}
			else
			{
				//Decrement address
				gAccAddress--;
			}
			
			//Display Acc No.
			SegDisplay.SetAccessory(gAccAddress);
			
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F0))
		{
			//Diverse
			
			if( gAccAddress == 0)
			{
				return;
			}
			
			//DCC mode
			
			Gateway.SetTurnout(gAccProtcol + gAccAddress, 0);
			
			//Display Message
			SegDisplay.SetWord(SSEG_D, SSEG_I, SSEG_V, SSEG_MINUS);
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F1))
		{
			//Straight
			
			if( gAccAddress == 0)
			{
				return;
			}
			
			Gateway.SetTurnout(gAccProtcol + gAccAddress, 1);
			
			//Display Message
			SegDisplay.SetWord(SSEG_S, SSEG_T, SSEG_R, SSEG_I);
			gDisplayWordCount = WAIT_MSG;
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_F2))
		{
			//Go to Loc mode
			
			gMode_Drive = 0;
			
			//Display Loc No.
			SegDisplay.SetNumber(gLocAddress);
		}
		else if(checkButton(gButton_val, aButton_val, BUTTON_DIR))
		{
			
			gLocDirection = (gLocDirection == GO_FWD) ? GO_REV : GO_FWD;
			
			Gateway.SetLocoDirection(gLocProtcol + gLocAddress, gLocDirection);
			
			//Display Message
			if( gLocDirection == GO_FWD)
			{
				SegDisplay.SetWord(SSEG_NONE, SSEG_F, SSEG_W, SSEG_D);
			}
			else
			{
				SegDisplay.SetWord(SSEG_NONE, SSEG_R, SSEG_E, SSEG_V);
			}
			gDisplayWordCount = WAIT_MSG;
			
			//Update EEPROM
			updateEEPROM();
		}
		
		
		
		//Store to buf
		gButton_val = aButton_val;
		
	}
	else
	{
		//Buttons have no difference from previous state
	}
	
}




/*******************************
    getButtonStatus
********************************/

byte getButtonStatus()
{
	
	byte aButton_val = 0;
	
	//Input Buttons
	bitWrite(aButton_val, 0, digitalRead(8)); // Down
	bitWrite(aButton_val, 1, digitalRead(7));// Up
	bitWrite(aButton_val, 2, digitalRead(11));// F0
	bitWrite(aButton_val, 3, digitalRead(12));// F1
	bitWrite(aButton_val, 4, digitalRead(13));// F2
	bitWrite(aButton_val, 5, (digitalRead(A1) == 1) ? 0 : 1);// DIR
	
	return ~aButton_val;
	
}

/*******************************
    checkButton
********************************/

bool checkButton(byte inBuf, byte inCurrent, byte inButtonBit)
{

	if( ((inBuf & inButtonBit) == 0) && ((inCurrent & inButtonBit) > 0))
	{
		return true;
	}
	else
	{
		return false;
	}
	
}

/*******************************
    checkButtonLeaved
********************************/

bool checkButtonLeaved(byte inBuf, byte inCurrent, byte inButtonBit)
{

	if( ((inBuf & inButtonBit) > 0) && ((inCurrent & inButtonBit) == 0))
	{
		return true;
	}
	else
	{
		return false;
	}
	
}

/*******************************
    checkButtonContinuous
********************************/

bool checkButtonContinuous(byte inBuf, byte inCurrent, byte inButtonBit)
{

	if( (inCurrent & inButtonBit) > 0)
	{
		return true;
	}
	else
	{
		return false;
	}
	
}

/*******************************
    Update EEPROM
********************************/

void updateEEPROM()
{
	if( gLocAddress != eepLocAddress)
	{
		//Update Loc Addr EEPROM
		saveEEPROM(0);
		
		//Update buffer
		eepLocAddress = gLocAddress;
	}
	
}

/*******************************
    Load EEPROM
********************************/

void loadEEPROM()
{
	
	if( EEPROM.read(0) == 0xCC)
	{
		
		eepLocAddress = (EEPROM.read(3) << 8) + EEPROM.read(2);
		gLocAddress = eepLocAddress;
		
	}
	
}

/*******************************
    Save EEPROM
********************************/

void saveEEPROM(byte inMode)
{
	
	if( EEPROM.read(0) == 0xCC)
	{
		
		if( inMode == 0)
		{
			EEPROM.write(2, lowByte(gLocAddress));
			EEPROM.write(3, highByte(gLocAddress));
			
		}
	}
	else
	{
		//Mint can ID
		EEPROM.write(0, 0xCC);
		
		//Data write
		EEPROM.write(1, 0);
		EEPROM.write(2, lowByte(gLocAddress));
		EEPROM.write(3, highByte(gLocAddress));
		
	}
	
}



